@@ -1,21 +1,19 @@
 #!/usr/bin/env python
 
 import cmd
-import readline
-import sys
-import traceback
-import time
-import threading
+import logging
 import os
+import sys
+import threading
+import time
 
-import spotify
-from spotify.manager import SpotifySessionManager, SpotifyPlaylistManager, \
-    SpotifyContainerManager
-try:
-    from spotify.alsahelper import AlsaController
-except ImportError:
-    from spotify.osshelper import OssController as AlsaController
-from spotify import Link, SpotifyError, ToplistBrowser
+from spotify import ArtistBrowser, Link, ToplistBrowser
+from spotify.audiosink import import_audio_sink
+from spotify.manager import (SpotifySessionManager, SpotifyPlaylistManager,
+    SpotifyContainerManager)
+
+AudioSink = import_audio_sink()
+container_loaded = threading.Event()
 
 class JukeboxUI(cmd.Cmd, threading.Thread):
 
@@ -30,14 +28,20 @@ class JukeboxUI(cmd.Cmd, threading.Thread):
         self.results = False
 
     def run(self):
-        self.cmdloop()
+        container_loaded.wait()
+        container_loaded.clear()
+        try:
+            self.cmdloop()
+        finally:
+            self.do_quit(None)
 
     def do_logout(self, line):
         self.jukebox.session.logout()
 
     def do_quit(self, line):
+        self.jukebox.stop()
+        self.jukebox.disconnect()
         print "Goodbye!"
-        self.jukebox.terminate()
         return True
 
     def do_list(self, line):
@@ -66,10 +70,19 @@ class JukeboxUI(cmd.Cmd, threading.Thread):
                 playlist = self.jukebox.starred
             for i, t in enumerate(playlist):
                 if t.is_loaded():
-                    print "%3d %s - %s" % (i, t.artists()[0].name(), t.name())
+                    print "%3d %s - %s [%s]" % (
+                        i, t.artists()[0].name(), t.name(),
+                        self.pretty_duration(t.duration()))
                 else:
                     print "%3d %s" % (i, "loading...")
 
+    def pretty_duration(self, milliseconds):
+        seconds = milliseconds // 1000
+        minutes = seconds // 60
+        seconds = seconds % 60
+        duration = '%02d:%02d' % (minutes, seconds)
+        return duration
+
     def do_play(self, line):
         if not line:
             self.jukebox.play()
@@ -84,10 +97,15 @@ class JukeboxUI(cmd.Cmd, threading.Thread):
         else:
             try:
                 playlist, track = map(int, line.split(' ', 1))
+                self.jukebox.load(playlist, track)
             except ValueError:
-                print "Usage: play [track_link] | [playlist] [track]"
-                return
-            self.jukebox.load(playlist, track)
+                try:
+                    playlist = int(line)
+                    self.jukebox.load_playlist(playlist)
+                except ValueError:
+                    print("Usage: play [track_link] | "
+                          "[playlist] [track] | [playlist]")
+                    return
         self.jukebox.play()
 
     def do_browse(self, line):
@@ -98,10 +116,23 @@ class JukeboxUI(cmd.Cmd, threading.Thread):
         if not l.type() in [Link.LINK_ALBUM, Link.LINK_ARTIST]:
             print "You can only browse albums and artists"
             return
-        def browse_finished(browser):
-            print "Browse finished"
+        def browse_finished(browser, userdata):
+            print "Browse finished, %s" % (userdata)
         self.jukebox.browse(l, browse_finished)
 
+    def print_search_results(self):
+        print "Artists:"
+        for a in self.results.artists():
+            print "    ", Link.from_artist(a), a.name()
+        print "Albums:"
+        for a in self.results.albums():
+            print "    ", Link.from_album(a), a.name()
+        print "Tracks:"
+        for a in self.results.tracks():
+            print "    ", Link.from_track(a, 0), a.name()
+        print self.results.total_tracks() - len(self.results.tracks()), \
+            "Tracks not shown"
+
     def do_search(self, line):
         if not line:
             if self.results is False:
@@ -118,12 +149,15 @@ class JukeboxUI(cmd.Cmd, threading.Thread):
                 print "Tracks:"
                 for a in self.results.tracks():
                     print "    ", Link.from_track(a, 0), a.name()
-                print self.results.total_tracks() - len(self.results.tracks()), "Tracks not shown"
+                print self.results.total_tracks() - \
+                        len(self.results.tracks()), "Tracks not shown"
         else:
+            line = line.decode('utf-8')
             self.results = None
             def search_finished(results, userdata):
                 print "\nSearch results received"
                 self.results = results
+                self.print_search_results()
             self.jukebox.search(line, search_finished)
 
     def do_queue(self, line):
@@ -150,7 +184,8 @@ class JukeboxUI(cmd.Cmd, threading.Thread):
     def do_watch(self, line):
         if not line:
             print """Usage: watch [playlist]
-You will be notified when tracks are added, moved or removed from the playlist."""
+You will be notified when tracks are added, moved or removed from the
+playlist."""
         else:
             try:
                 p = int(line)
@@ -177,7 +212,7 @@ You will be notified when tracks are added, moved or removed from the playlist."
             self.jukebox.watch(self.jukebox.ctr[p], True)
 
     def do_toplist(self, line):
-        usage = "Usage: toplist (albums|artists|tracks) (GB|FR|..|NO|all|current)"
+        usage = "Usage: toplist (albums|artists|tracks) (GB|FR|..|all|current)"
         if not line:
             print usage
         else:
@@ -194,7 +229,8 @@ You will be notified when tracks are added, moved or removed from the playlist."
         if not line:
             print "Usage: add_new_playlist <name>"
         else:
-          new_playlist = self.jukebox.ctr.add_new_playlist(line.decode('utf-8'))
+            new_playlist = self.jukebox.ctr.add_new_playlist(
+                line.decode('utf-8'))
 
     def do_add_to_playlist(self, line):
         usage = "Usage: add_to_playlist <playlist_index> <insert_point>" + \
@@ -215,9 +251,12 @@ You will be notified when tracks are added, moved or removed from the playlist."
                 tracks = self.results.tracks()
                 for i in args:
                     for a in tracks[int(i)].artists():
-                        print u'{}. {} - {} '.format(i,a.name(),tracks[int(i)].name())
-                print u'adding them to {} '.format(self.jukebox.ctr[index].name())
-                self.jukebox.ctr[index].add_tracks(insert,[tracks[int(i)] for i in args])
+                        print u'{0}. {1} - {2} '.format(
+                            i, a.name(), tracks[int(i)].name())
+                print u'adding them to {0} '.format(
+                    self.jukebox.ctr[index].name())
+                self.jukebox.ctr[index].add_tracks(
+                    insert, [tracks[int(i)] for i in args])
 
     do_ls = do_list
     do_EOF = do_quit
@@ -237,7 +276,7 @@ class JukeboxPlaylistManager(SpotifyPlaylistManager):
 ## container calllbacks ##
 class JukeboxContainerManager(SpotifyContainerManager):
     def container_loaded(self, c, u):
-        print 'Container loaded !'
+        container_loaded.set()
 
     def playlist_added(self, c, p, i, u):
         print 'Container: playlist "%s" added.' % p.name()
@@ -257,28 +296,28 @@ class Jukebox(SpotifySessionManager):
 
     def __init__(self, *a, **kw):
         SpotifySessionManager.__init__(self, *a, **kw)
-        self.audio = AlsaController()
+        self.audio = AudioSink(backend=self)
         self.ui = JukeboxUI(self)
         self.ctr = None
         self.playing = False
         self._queue = []
         self.playlist_manager = JukeboxPlaylistManager()
         self.container_manager = JukeboxContainerManager()
+        self.track_playing = None
         print "Logging in, please wait..."
 
+    def new_track_playing(self, track):
+        self.track_playing = track
 
     def logged_in(self, session, error):
         if error:
             print error
             return
         self.session = session
-        try:
-            self.ctr = session.playlist_container()
-            self.container_manager.watch(self.ctr)
-            self.starred = session.starred()
-            self.ui.start()
-        except:
-            traceback.print_exc()
+        self.ctr = session.playlist_container()
+        self.container_manager.watch(self.ctr)
+        self.starred = session.starred()
+        self.ui.start()
 
     def logged_out(self, session):
         self.ui.cmdqueue.append("quit")
@@ -286,6 +325,7 @@ class Jukebox(SpotifySessionManager):
     def load_track(self, track):
         if self.playing:
             self.stop()
+        self.new_track_playing(track)
         self.session.load(track)
         print "Loading %s" % track.name()
 
@@ -296,17 +336,38 @@ class Jukebox(SpotifySessionManager):
             pl = self.ctr[playlist]
         elif playlist == len(self.ctr):
             pl = self.starred
-        self.session.load(pl[track])
-        print "Loading %s from %s" % (pl[track].name(), pl.name())
+        spot_track = pl[track]
+        self.new_track_playing(spot_track)
+        self.session.load(spot_track)
+        print "Loading %s from %s" % (spot_track.name(), pl.name())
+
+    def load_playlist(self, playlist):
+        if self.playing:
+            self.stop()
+        if 0 <= playlist < len(self.ctr):
+            pl = self.ctr[playlist]
+        elif playlist == len(self.ctr):
+            pl = self.starred
+        print "Loading playlist %s" % pl.name()
+        if len(pl):
+            print "Loading %s from %s" % (pl[0].name(), pl.name())
+            self.new_track_playing(pl[0])
+            self.session.load(pl[0])
+        for i, track in enumerate(pl):
+            if i == 0:
+                continue
+            self._queue.append((playlist, i))
 
     def queue(self, playlist, track):
         if self.playing:
             self._queue.append((playlist, track))
         else:
+            print 'Loading %s', track.name()
             self.load(playlist, track)
             self.play()
 
     def play(self):
+        self.audio.start()
         self.session.play(1)
         print "Playing"
         self.playing = True
@@ -315,22 +376,22 @@ class Jukebox(SpotifySessionManager):
         self.session.play(0)
         print "Stopping"
         self.playing = False
+        self.audio.stop()
 
-    def music_delivery(self, *a, **kw):
-        return self.audio.music_delivery(*a, **kw)
+    def music_delivery_safe(self, *args, **kwargs):
+        return self.audio.music_delivery(*args, **kwargs)
 
     def next(self):
         self.stop()
         if self._queue:
-            t = self._queue.pop()
+            t = self._queue.pop(0)
             self.load(*t)
             self.play()
         else:
             self.stop()
 
     def end_of_track(self, sess):
-        print "track ends."
-        self.next()
+        self.audio.end_of_track()
 
     def search(self, *args, **kwargs):
         self.session.search(*args, **kwargs)
@@ -341,19 +402,18 @@ class Jukebox(SpotifySessionManager):
             while not browser.is_loaded():
                 time.sleep(0.1)
             for track in browser:
-                print track
+                print track.name()
         if link.type() == link.LINK_ARTIST:
-            browser = self.session.browse_artist(link.as_artist(), callback)
+            browser = ArtistBrowser(link.as_artist())
             while not browser.is_loaded():
                 time.sleep(0.1)
             for album in browser:
                 print album.name()
-        callback(browser)
 
     def watch(self, p, unwatch=False):
         if not unwatch:
             print "Watching playlist: %s" % p.name()
-            self.playlist_manager.watch(p);
+            self.playlist_manager.watch(p)
         else:
             print "Unatching playlist: %s" % p.name()
             self.playlist_manager.unwatch(p)
@@ -375,8 +435,12 @@ class Jukebox(SpotifySessionManager):
 if __name__ == '__main__':
     import optparse
     op = optparse.OptionParser(version="%prog 0.1")
-    op.add_option("-u", "--username", help="spotify username")
-    op.add_option("-p", "--password", help="spotify password")
+    op.add_option("-u", "--username", help="Spotify username")
+    op.add_option("-p", "--password", help="Spotify password")
+    op.add_option("-v", "--verbose", help="Show debug information",
+        dest="verbose", action="store_true")
     (options, args) = op.parse_args()
+    if options.verbose:
+        logging.basicConfig(level=logging.DEBUG)
     session_m = Jukebox(options.username, options.password, True)
     session_m.connect()
